from osc_kreuz.renderer import Renderer, ViewClient
import osc_kreuz.str_keys_conventions as skc
from oscpy.server import OSCThreadServer
from functools import partial
from osc_kreuz.soundobject import SoundObject
import ipaddress
import logging

log = logging.getLogger("OSCcomcenter")


soundobjects: list[SoundObject] = []

clientSubscriptions = {}
receivers: list[Renderer] = []
globalconfig = dict()
extendedOscInput = True
verbosity = 0
bPrintOSC = False


def setVerbosity(v: int):
    global verbosity, bPrintOSC
    verbosity = v
    bPrintOSC = v >= 2
    Renderer.setVerbosity(v)
    log.debug("verbosity set to", v)


osc_ui_server = OSCThreadServer()
osc_data_server = OSCThreadServer()
osc_setting_server = OSCThreadServer()


oscr_ip = False


def setupOscSettingsBindings():
    osc_setting_server.listen(
        address="0.0.0.0", port=globalconfig["inputport_settings"], default=True
    )

    osc_setting_server.bind(
        "/oscrouter/debug/osccopy".encode(), oscreceived_debugOscCopy
    )
    osc_setting_server.bind("/oscrouter/debug/verbose".encode(), oscreceived_verbose)
    osc_setting_server.bind(
        "/oscrouter/subscribe".encode(), oscreceived_subscriptionRequest
    )
    osc_setting_server.bind("/oscrouter/ping".encode(), oscreceived_ping)
    osc_setting_server.bind("/oscrouter/pong".encode(), oscreceived_pong)
    osc_setting_server.bind("/oscrouter/dump".encode(), oscreceived_dump)

    global oscr_ip
    if "oscr_ip" in globalconfig.keys() and checkIp(globalconfig["oscr_ip"]):
        oscr_ip = globalconfig["oscr_ip"].encode()


def oscreceived_ping(*args):

    if checkPort(args[0]):
        vals = [oscr_ip] if oscr_ip else []
        osc_setting_server.answer(b"/oscrouter/pong", port=args[0], values=vals)


def oscreceived_pong(*args):

    try:
        clientName = args[0]
        clientSubscriptions[clientName].receivedIsAlive()
    except:
        if verbosity > 0:
            _name = ""
            if len(args) > 0:
                _name = args[0]
            log.info("no renderer for pong message {}".format(_name))


def oscreceived_subscriptionRequest(*args) -> None:
    """OSC Callback for subscription Requests.

    These requests follow the format:
    /oscrouter/subscribe myname 31441 xyz 0 5
    /oscrouter/subscribe [client_name] [client_port] [coordinate_format] [source index as value? (0 or 1)] [update rate]
    args[0] nameFor Client
    args[1] port client listens to
    args[2] format client expects
    args[3] send source index as value instead of inside the osc prefix
    args[4] source position update rate
    """

    viewClientInitValues = {}
    vCName = args[0]
    subArgs = len(args)
    if subArgs >= 2:
        if checkPort(args[1]):
            viewClientInitValues["port"] = args[1]

            _ip = osc_setting_server.get_sender()[1]

            viewClientInitValues["ip_address"] = _ip

            # if subArgs>2:
            #     initKeys = ['dataformat', 'indexAsValue', 'updateintervall']
            #     for i in range(2, subArgs):
            #         viewClientInitValues[initKeys[i-2]] = args[i]
            try:
                viewClientInitValues["dataformat"] = args[2].decode()
            except:
                pass
            try:
                viewClientInitValues["indexAsValue"] = args[3]
            except:
                pass
            try:
                viewClientInitValues["updateintervall"] = args[4]
            except:
                pass

        newViewClient = ViewClient(vCName, **viewClientInitValues)

        clientSubscriptions[vCName] = newViewClient
        # TODO check if this is threadsafe (it probably isn't)
        receivers.append(newViewClient)
        newViewClient.checkAlive(deleteClient)

    else:
        if verbosity > 0:
            log.info("not enough arguments für view client")


def oscreceived_dump(*args):
    pass
    # TODO: dump all source data to renderer


def deleteClient(viewC, alias):
    # TODO check if this is threadsafe (it probably isn't)

    if verbosity > 0:
        log.info("deleting client", viewC, alias)
    receivers.remove(viewC)
    del clientSubscriptions[alias]


def checkPort(port) -> bool:
    if type(port) == int and 1023 < port < 65535:
        return True
    else:
        if verbosity > 0:
            log.info("port", port, "not legit")
        return False


def checkIp(ip) -> bool:
    ipalright = True
    try:
        _ip = "127.0.0.1" if ip == "localhost" else ip
        _ = ipaddress.ip_address(_ip)
    except:
        ipalright = False
        if verbosity > 0:
            log.info("ip address", ip, "not legit")

    return ipalright


def checkIpAndPort(ip, port) -> bool:
    return checkIp(ip) and checkPort(port)


def oscreceived_debugOscCopy(*args):
    ip = ""
    port = 0
    if len(args) == 2:
        ip = args[0].decode()
        port = args[1]
    elif len(args) == 1:
        ipport = args[0].decode().split(":")
        if len(ipport) == 2:
            ip = ipport[0]
            port = ipport[1]
    else:
        Renderer.debugCopy = False
        log.info("debug client: invalid message format")
        return
    try:
        ip = "127.0.0.1" if ip == "localhost" else ip
        osccopy_ip = ipaddress.ip_address(ip)
        osccopy_port = int(port)
    except:
        log.info("debug client: invalid ip or port")
        return
    log.info(f"debug client connected: {ip}:{port}")

    if 1023 < osccopy_port < 65535:
        Renderer.createDebugClient(str(osccopy_ip), osccopy_port)
        Renderer.debugCopy = True
        return

    Renderer.debugCopy = False


def oscreceived_verbose(*args):
    vvvv = -1
    try:
        vvvv = int(args[0])
    except:
        setVerbosity(0)
        # verbosity = 0
        # Renderer.setVerbosity(0)
        log.error("wrong verbosity argument")
        return

    if 0 <= vvvv <= 2:
        setVerbosity(vvvv)
        # global verbosity
        # verbosity = vvvv
        # Renderer.setVerbosity(vvvv)
    else:
        setVerbosity(0)


def build_osc_paths(
    osc_path_type: skc.OscPathType, value: str, idx: int | None = None
) -> list[str]:
    """Builds a list of all needed osc paths for a given osc path Type and the value.
    If idx is supplied, the extended path is used. Aliases for the value are handled

    Args:
        osc_path_type (skc.OscPathType): Osc Path Type
        value (str): value to be written into the OSC strings.
        idx (int | None, optional): Index of the source if the extended format should be used. Defaults to None.

    Raises:
        KeyError: Raised when the Osc Path Type does not exist

    Returns:
        list[str]: list of OSC path strings
    """
    if osc_path_type not in skc.osc_paths:
        raise KeyError(f"Invalid OSC Path Type: {osc_path_type}")
    try:
        aliases = skc.osc_aliases[value]
    except KeyError:
        aliases = [value]

    if idx is None:
        paths = skc.osc_paths[osc_path_type]["base"]
    else:
        paths = skc.osc_paths[osc_path_type]["extended"]

    return [path.format(val=alias, idx=idx) for alias in aliases for path in paths]


def setupOscBindings():
    """Sets up all Osc Bindings"""
    setupOscSettingsBindings()

    osc_ui_server.listen(
        address="0.0.0.0", port=globalconfig[skc.inputport_ui], default=True
    )
    osc_data_server.listen(
        address="0.0.0.0", port=globalconfig[skc.inputport_data], default=True
    )

    n_sources = globalconfig["number_sources"]

    # Setup OSC Callbacks for positional data
    for key in skc.posformat.keys():

        for addr in build_osc_paths(skc.OscPathType.Position, key):
            bindToDataAndUiPort(addr, partial(oscreceived_setPosition, key))

        if extendedOscInput:
            for i in range(n_sources):
                idx = i + 1
                for addr in build_osc_paths(skc.OscPathType.Position, key, idx=idx):
                    bindToDataAndUiPort(
                        addr, partial(oscreceived_setPositionForSource, key, i)
                    )

    # Setup OSC for Wonder Attribute Paths
    for key in skc.SourceAttributes:
        for addr in build_osc_paths(skc.OscPathType.Properties, key.value):
            log.info(f"WFS Attr path: {addr}")
            bindToDataAndUiPort(addr, partial(oscReceived_setValueForAttribute, key))

        for i in range(n_sources):
            idx = i + 1
            for addr in build_osc_paths(skc.OscPathType.Properties, key.value, idx):
                bindToDataAndUiPort(
                    addr, partial(oscreceived_setValueForSourceForAttribute, i, key)
                )

    # sendgain input
    for spatGAdd in ["/source/send/spatial", "/send/gain", "/source/send"]:
        bindToDataAndUiPort(spatGAdd, partial(oscreceived_setRenderGain))

    # get list of all render units
    try:
        render_units = globalconfig["render_units"]
    except KeyError:
        render_units = []

    # Setup OSC Callbacks for all render units
    for rendIdx, render_unit in enumerate(render_units):
        # get aliases for this render unit, if none exist just use the base name

        # add callback to base paths for all all aliases
        for addr in build_osc_paths(skc.OscPathType.Gain, render_unit):
            bindToDataAndUiPort(
                addr, partial(oscreceived_setRenderGainToRenderer, rendIdx)
            )

        # add callback to extended paths
        if extendedOscInput:
            for i in range(n_sources):
                idx = i + 1
                for addr in build_osc_paths(skc.OscPathType.Gain, render_unit, idx):
                    bindToDataAndUiPort(
                        addr,
                        partial(
                            oscreceived_setRenderGainForSourceForRenderer, i, rendIdx
                        ),
                    )

    directSendAddr = "/source/send/direct"
    bindToDataAndUiPort(directSendAddr, partial(oscreceived_setDirectSend))

    # XXX can this be removed?
    # if extendedOscInput:
    #     for i in range(n_sources):
    #         idx = i + 1
    #         for addr in [
    #             ("/source/" + str(idx) + "/rendergain"),
    #             ("/source/" + str(idx) + "/send/spatial"),
    #             ("/source/" + str(idx) + "/spatial"),
    #             ("/source/" + str(idx) + "/sendspatial"),
    #         ]:

    #             bindToDataAndUiPort(
    #                 addr, partial(oscreceived_setRenderGainForSource, i)
    #             )

    #             # TODO fix whatever this is
    #             # This adds additional osc paths for the render engines by index
    #             # for j in range(len(renderengineClients)):
    #             #     addr2 = addr + "/" + str(j)
    #             #     bindToDataAndUiPort(
    #             #         addr2,
    #             #         partial(oscreceived_setRenderGainForSourceForRenderer, i, j),
    #             #     )

    #         for addr in [
    #             ("/source/" + str(idx) + "/direct"),
    #             ("/source/" + str(idx) + "/directsend"),
    #             ("/source/" + str(idx) + "/senddirect"),
    #             ("/source/" + str(idx) + "/send/direct"),
    #         ]:
    #             bindToDataAndUiPort(
    #                 addr, partial(oscreceived_setDirectSendForSource, idx)
    #             )

    #             for j in range(globalconfig["number_direct_sends"]):
    #                 addr2 = addr + "/" + str(j)
    #                 bindToDataAndUiPort(
    #                     addr2,
    #                     partial(oscreceived_setDirectSendForSourceForChannel, idx, j),
    #                 )

    if verbosity > 2:
        for add in osc_ui_server.addresses:
            log.info(add)


def bindToDataAndUiPort(addr: str, func):
    # dontUseDataPortFlag = bool(globalconfig['data_port_timeout'] == 0)
    log.debug(f"Adding OSC callback for {addr}")
    addrEnc = addr.encode()

    # if verbosity >= 2:
    osc_ui_server.bind(
        addrEnc, partial(printOSC, addr=addr, port=globalconfig[skc.inputport_ui])
    )
    osc_data_server.bind(
        addrEnc, partial(printOSC, addr=addr, port=globalconfig[skc.inputport_data])
    )

    osc_ui_server.bind(addrEnc, partial(func, fromUi=True))
    osc_data_server.bind(addrEnc, partial(func, fromUi=False))


def sourceLegit(id: int) -> bool:
    indexInRange = id in range(globalconfig["number_sources"])
    if verbosity > 0:
        if not indexInRange:
            if not type(id) == int:
                log.warn("source index is no integer")
            else:
                log.warn("source index out of range")
    return indexInRange


def renderIndexLegit(id: int) -> bool:
    indexInRange = id in range(globalconfig["n_renderengines"])
    if verbosity > 0:
        if not indexInRange:
            if not type(id) == int:
                log.warn("renderengine index is no integer")
            else:
                log.warn("renderengine index out of range")
    return indexInRange


def directSendLegit(id: int) -> bool:
    indexInRange = id in range(globalconfig["number_direct_sends"])
    if verbosity > 0:
        if not indexInRange:
            if not type(id) == int:
                log.warn("direct send index is no integer")
            else:
                log.warn("direct send index out of range")
    return indexInRange


def oscreceived_setPosition(coordKey, *args, fromUi=True):
    sIdx = args[0] - 1
    if sourceLegit(sIdx):
        sIdx = int(sIdx)
        oscreceived_setPositionForSource(coordKey, sIdx, *args[1:], fromUi=fromUi)


def oscreceived_setPositionForSource(coordKey, sIdx: int, *args, fromUi=True):

    if soundobjects[sIdx].setPosition(coordKey, *args, fromUi=fromUi):
        notifyRenderClientsForUpdate("sourcePositionChanged", sIdx, fromUi=fromUi)
        # notifyRendererForSourcePosition(sIdx, fromUi)


# TODO why are there this many functions for doing almost the same thing?
def oscreceived_setRenderGain(*args, fromUi: bool = True):
    sIdx = args[0] - 1
    if sourceLegit(sIdx):
        sIdx = int(sIdx)
        oscreceived_setRenderGainForSource(sIdx, *args[1:], fromUi)


def oscreceived_setRenderGainToRenderer(rIdx: int, *args, fromUi: bool = True):
    sIdx = args[0] - 1
    if renderIndexLegit(rIdx) and sourceLegit(sIdx):
        rIdx = int(rIdx)
        sIdx = int(sIdx)
        oscreceived_setRenderGainForSourceForRenderer(
            sIdx, rIdx, *args[1:], fromUi=fromUi
        )


def oscreceived_setRenderGainForSource(sIdx: int, *args, fromUi: bool = True):
    rIdx = args[0]
    if renderIndexLegit(rIdx):
        rIdx = int(rIdx)
        oscreceived_setRenderGainForSourceForRenderer(
            sIdx, rIdx, *args[1:], fromUi=fromUi
        )


def oscreceived_setRenderGainForSourceForRenderer(
    sIdx: int, rIdx: int, *args, fromUi: bool = True
):

    if soundobjects[sIdx].setRendererGain(rIdx, args[0], fromUi):
        notifyRenderClientsForUpdate(
            "sourceRenderGainChanged", sIdx, rIdx, fromUi=fromUi
        )


def oscreceived_setDirectSend(*args, fromUi: bool = True):
    sIdx = args[0] - 1
    if sourceLegit(sIdx):
        sIdx = int(sIdx)
        oscreceived_setDirectSendForSource(sIdx, *args[1:], fromUi)


def oscreceived_setDirectSendForSource(sIdx: int, *args, fromUi: bool = True):
    cIdx = args[0]
    if directSendLegit(cIdx):  # 0 <= cIdx < globalconfig['number_direct_sends']:
        cIdx = int(cIdx)
        oscreceived_setDirectSendForSourceForChannel(sIdx, cIdx, *args[1:], fromUi)


def oscreceived_setDirectSendForSourceForChannel(
    sIdx: int, cIdx: int, *args, fromUi: bool = True
):
    if soundobjects[sIdx].setDirectSend(cIdx, args[0], fromUi):
        notifyRenderClientsForUpdate(
            "sourceDirectSendChanged", sIdx, cIdx, fromUi=fromUi
        )


# TODO: implement this thing
def notifyRendererForDirectsendGain(sIdx: int, cIfx: int, fromUi: bool = True):
    pass


def oscreceived_setAttribute(*args, fromUi: bool = True):
    sIdx = args[0] - 1
    if sourceLegit(sIdx):
        sIdx = int(sIdx)
        oscreceived_setAttributeForSource(sIdx, *args[1:], fromUi)


def oscreceived_setAttributeForSource(sIdx: int, *args, fromUi: bool = True):
    attribute = args[0]
    if attribute in skc.knownAttributes:
        oscreceived_setValueForSourceForAttribute(sIdx, attribute, *args[1:], fromUi)


def oscReceived_setValueForAttribute(
    attribute: skc.SourceAttributes, *args, fromUi: bool = True
):
    sIdx = args[0] - 1
    if sourceLegit(sIdx):
        sIdx = int(sIdx)
        oscreceived_setValueForSourceForAttribute(sIdx, attribute, *args[1:], fromUi)


def oscreceived_setValueForSourceForAttribute(
    sIdx: int, attribute: skc.SourceAttributes, *args, fromUi: bool = True
):
    if soundobjects[sIdx].setAttribute(attribute, args[0], fromUi):
        notifyRenderClientsForUpdate(
            "sourceAttributeChanged", sIdx, attribute, fromUi=fromUi
        )


def notifyRenderClientsForUpdate(updateFunction: str, *args, fromUi: bool = True):
    for receiver in receivers:
        updatFunc = getattr(receiver, updateFunction)
        updatFunc(*args)

    # XXX why was this distinction made? dataClients were not included in receivers
    # if fromUi:
    #     for rend in dataClients:
    #         updatFunc = getattr(rend, updateFunction)
    #         updatFunc(*args)


######
def oscreceived_sourceAttribute(attribute: skc.SourceAttributes, *args):

    sidx = int(args[0]) - 1
    if sidx >= 0 and sidx < 64:
        oscreceived_sourceAttribute_wString(sidx, attribute, args[1:])


def oscreceived_sourceAttribute_wString(
    sidx: int, attribute: skc.SourceAttributes, *args
):
    sobject = soundobjects[sidx]
    if sobject.setAttribute(attribute, args[0]):
        for ren in receivers:
            ren.sourceAttributeChanged(sidx, attribute)


def printOSC(*args, addr: str = "", port: int = 0):
    if bPrintOSC:
        log.info("incoming OSC on Port", port, addr, args)



from osc_kreuz.renderer import (
    Update,
    GainUpdate,
    PositionUpdate,
    AttributeUpdate,
    DirectSendUpdate,
    wonderPlanewaveAttributeUpdate,
)

import osc_kreuz.renderer as r
from osc_kreuz.soundobject import SoundObject
import osc_kreuz.str_keys_conventions as skc


def test_update_classes():
    updates = set()
    u1 = Update(b"/", None)
    u2 = Update(b"/", None)

    for u in [u1, u2]:
        updates.add(u)

    assert len(updates) == 1

    for u in [
        GainUpdate(b"/", None, render_idx=1, source_index=1),
        GainUpdate(b"/", None, render_idx=1, source_index=2),
        GainUpdate(b"/", None, render_idx=2, source_index=2),
        GainUpdate(b"/", None, render_idx=2, source_index=2),
    ]:
        updates.add(u)
    assert len(updates) == 4

    for u in [
        PositionUpdate(b"/", None, coord_fmt=None, source_index=5),
        PositionUpdate(b"/", None, coord_fmt=None, source_index=5),
        PositionUpdate(b"/", None, coord_fmt=None, source_index=6),
        PositionUpdate(b"/", None, coord_fmt=None, source_index=6),
    ]:
        updates.add(u)
    assert len(updates) == 6

    for u in [
        AttributeUpdate("/", None, None),
        AttributeUpdate("/", None, None),
    ]:
        updates.add(u)

    assert len(updates) == 7

    for u in [
        DirectSendUpdate("/", None, None, source_index=1),
        DirectSendUpdate("/", None, None),
    ]:
        updates.add(u)

    assert len(updates) == 9

    for u in [
        wonderPlanewaveAttributeUpdate("/", None, None, source_index=1),
        wonderPlanewaveAttributeUpdate("/", None, None),
    ]:
        updates.add(u)

    assert len(updates) == 11


def prepare_renderer(conf: dict, disable_network: bool = True):

    global_conf = {
        "number_direct_sends": 2,
        "data_port_timeout": 0,
        "render_units": ["ambi", "wfs", "reverb"],
    }

    def function_override(*args, **kwargs):
        pass

    SoundObject.readGlobalConfig(global_conf)

    r.Renderer.numberOfSources = 1
    r.Renderer.sources = [SoundObject(), SoundObject()]
    r.Renderer.globalConfig = global_conf
    if disable_network:
        r.Renderer.scheduleSourceUpdateCheck = function_override
        r.Renderer.updateSource = function_override
    client = r.createRendererClient(conf)
    return client


def check_source_update(
    renderer: r.Renderer,
    update_type: str,
    source_idx,
    expected_path: list[bytes] | bytes,
    expected_output: list | list[list],
    param=None,
):
    if isinstance(expected_path, bytes):
        expected_path = [expected_path]
        expected_output = [expected_output]

    s = renderer.updateStack[source_idx]

    n_expected = len(expected_path)
    assert len(s) == 0

    match update_type:
        case "pos":
            renderer.sourcePositionChanged(source_idx)
        case "attr":
            renderer.sourceAttributeChanged(source_idx, param)
        case "gain":
            renderer.sourceRenderGainChanged(source_idx, param)
        case "send":
            renderer.sourceDirectSendChanged(source_idx, param)

    assert len(s) == n_expected

    for i in range(n_expected):
        u = s.pop()
        m = u.to_message()

        assert m.path in expected_path
        assert m.values in expected_output
        expected_path.remove(m.path)
        expected_output.remove(m.values)

    assert len(s) == 0


def test_wonder_renderer():
    conf = {"type": "Wonder", "hosts": [], "updateintervall": 5}
    c = prepare_renderer(conf)
    s = c.updateStack[0]
    so = c.sources[0]
    assert len(s) == 0
    assert isinstance(c, r.Wonder)

    check_source_update(
        c,
        "pos",
        0,
        b"/WONDER/source/position",
        [0, *so.getPosition(skc.CoordFormats.xy), c.interpolTime],
    )

    for attr, path, expected in [
        (skc.SourceAttributes.doppler, b"/WONDER/source/dopplerEffect", [0, 0]),
        (skc.SourceAttributes.planewave, b"/WONDER/source/type", [0, 1]),
        (
            skc.SourceAttributes.angle,
            b"/WONDER/source/angle",
            [0, so.getAttribute(skc.SourceAttributes.angle), c.interpolTime],
        ),
    ]:
        check_source_update(c, "attr", 0, path, expected, attr)


def test_audiomatrix_renderer():
    conf = {
        "type": "audiomatrix",
        "hosts": [],
        "paths": [
            {"path": "/source/gain/wfs", "renderer": "wfs", "type": "gain"},
            {"path": "/source/gain/ambi", "type": "gain", "renderer": "ambi"},
            {"path": "/source/pos", "type": "position", "format": "aed"},
            {"path": "/source/xyz", "type": "position", "format": "xyz"},
        ],
    }
    c = prepare_renderer(conf)
    s = c.updateStack[0]
    so = c.sources[0]
    assert len(s) == 0
    assert isinstance(c, r.AudioMatrix)

    check_source_update(
        c,
        "pos",
        0,
        [b"/source/pos", b"/source/xyz"],
        [
            [0, *so.getPosition(skc.CoordFormats.aed)],
            [0, *so.getPosition(skc.CoordFormats.xyz)],
        ],
    )

    for r_idx, path, expected in [
        (0, b"/source/gain/ambi", [0, 0]),
        (1, b"/source/gain/wfs", [0, 0]),
        (2, [], []),
    ]:
        check_source_update(c, "gain", 0, path, expected, r_idx)


def test_audiorouter_renderer():
    conf = {"type": "audiorouter", "hosts": [], "updateintervall": 5}
    c = prepare_renderer(conf)
    s = c.updateStack[0]
    so = c.sources[0]
    assert len(s) == 0
    assert isinstance(c, r.Audiorouter)

    check_source_update(
        c,
        "pos",
        0,
        [],
        [],
    )

    for r_idx, path, expected in [
        (0, b"/source/send/spatial", [0, 0, 0]),
        (1, [], []),
        (2, b"/source/reverb/gain", [0, 0]),
    ]:
        check_source_update(c, "gain", 0, path, expected, r_idx)


def test_audiorouterWFS_renderer():
    conf = {"type": "audiorouterWFS", "hosts": [], "updateintervall": 5}
    c = prepare_renderer(conf)
    s = c.updateStack[0]
    so = c.sources[0]
    assert len(s) == 0
    assert isinstance(c, r.AudiorouterWFS)

    check_source_update(
        c,
        "pos",
        0,
        [],
        [],
    )

    for r_idx, path, expected in [
        (0, [], []),
        (1, [b"/source/send/spatial"], [[0, 1, 0]]),
        (2, [], []),
    ]:
        check_source_update(c, "gain", 0, path, expected, r_idx)



# Comment on Conversions: In General x and y axis were swapped because 0° should be at the front which can easily achieved throu that

from typing import Any
import numpy as np


# change to f32 ?
def f32(val) -> np.float32:
    return np.float32(val)


deg90_rad = f32(np.deg2rad(90))


def deg2rad(angle_deg) -> np.float32:
    return f32(angle_deg * 0.01745329252)


def rad2deg(angle_rad) -> np.float32:
    return f32(angle_rad) * 57.2957795131


def mag_xyz(x, y, z) -> np.float32:
    return np.sqrt(np.square(x) + np.square(y) + np.square(z))


def convert_deg_angleToRad(azimuth, elevation) -> list[np.float32]:

    if azimuth < 0:
        azimuth += f32(360)

    azim_rad = np.deg2rad(azimuth)

    # if elevation < 0:
    #     elevation = f16(90) - elevation
    elev_rad = np.deg2rad(elevation)

    return [azim_rad, elev_rad]

    # self._positionIsSet[dk.polar_rad] = True


def convert_deg_angleToRad_correctMath(azimuth, elevation) -> list[np.float32]:

    _azi = wrapAzimuth180(azimuth - 90) * -1

    _aziRad = deg2rad(_azi)
    _elevRad = deg2rad(elevation)

    return [_aziRad, _elevRad]
    # return convert_deg_angleToRad(_azi, elevation)


def conv_pol2cart(azim, elev, dist) -> list[np.float32]:

    azim_rad, elev_rad = convert_deg_angleToRad(azim, elev)

    y = np.sin(deg90_rad - elev_rad) * np.cos(azim_rad) * dist
    x = np.sin(deg90_rad - elev_rad) * np.sin(azim_rad) * dist
    z = np.cos(deg90_rad - elev_rad) * dist

    return [x, y, z]

    # self._positionIsSet[dk.cartesian] = True


def conv_pol2ncart(azim, elev, dist) -> list[np.float32]:

    azim_rad, elev_rad = convert_deg_angleToRad(azim, elev)

    ny = np.sin(deg90_rad - elev_rad) * np.cos(azim_rad)
    nx = np.sin(deg90_rad - elev_rad) * np.sin(azim_rad)
    nz = np.cos(deg90_rad - elev_rad)

    return [nx, ny, nz, dist]

    # self._positionIsSet[skc.normcartesian] = True


def conv_cart2pol(x, y, z) -> list[np.float32]:
    dist = mag_xyz(x, y, z)
    azim_rad = np.arctan2(x, y)
    azim = np.rad2deg(azim_rad)
    if azim > 180.0:
        azim -= f32(360)

    elev_rad = np.arctan2(z, np.sqrt(np.square(x) + np.square(y)))
    elev = np.rad2deg(elev_rad)
    if elev > 90.0:
        elev = 90.0 - elev

    return [azim, elev, dist]


def conv_cart2ncart(x, y, z) -> list[np.float32]:
    nd = mag_xyz(x, y, z)
    inv_mag = 1 / nd
    nx = f32(x * inv_mag)
    ny = f32(y * inv_mag)
    nz = f32(z * inv_mag)
    return [nx, ny, nz, nd]


def conv_ncart2pol(nx, ny, nz, nd):
    azim_rad = np.arctan(nx / ny)
    azim = np.rad2deg(azim_rad)
    azim = wrapAzimuth180(azim)

    elev_rad = np.arctan2(nz, np.sqrt(np.square(nx) + np.square(ny)))
    elev = np.rad2deg(elev_rad)
    elev = wrapElevation90(elev)

    return [azim, elev, nd]


def conv_ncart2cart(nx, ny, nz, nd) -> list[np.float32]:
    x = nx * nd
    y = ny * nd
    z = nz * nd
    return [x, y, z]


def wrapAzimuth180(azim) -> np.float32:
    if azim > 180.0:
        azim -= f32(360)
        # return azim
    elif azim < -180.0:
        azim += 360.0

    while azim < -180.0 or azim > 180.0:
        azim = wrapAzimuth180(azim)

    return azim


def wrapElevation90(elev) -> np.float32:
    if elev > 90.0:
        elev = f32(90.0) - elev
        return elev
    else:
        return elev


# TODO: right conversions
def apply_factor_addition(values, factors, additions):
    # values
    return values


# TODO: implement this
def azi_to_wonderangle(azim) -> np.float32:
    return azim


def dist_to_gain(dist) -> np.float32:
    return dist


def convertedValue(x) -> Any:
    if isint(x):
        return int(x)
    elif isfloat(x):
        return float(x)
    else:
        return x


def isfloat(x):
    try:
        a = float(x)
    except (TypeError, ValueError):
        return False
    else:
        return True


def isint(x):
    try:
        a = float(x)
        b = int(a)
    except (TypeError, ValueError):
        return False
    else:
        return a == b



coord_format = "coord_format"
cartesian = "cartesian"
polar = "polar"
normcartesian = "normcartesian"

x = "x"
y = "y"
z = "z"
azim = "azim"
azimuth = "azim"
elev = "elev"
elevation = "elev"
dist = "dist"
distance = "dist"
angle = "angle"

az_rad = "az_r"
el_rad = "ev_r"
polar_rad = "pol_rad"


a = "a"
e = "e"
d = "d"

# normalised xyz on a sphere with r = 1
nx = "nx"
nxd = "nxd"
ny = "ny"
nyd = "nyd"
nz = "nz"
nzd = "nzd"
nd = "dist"  # distance

aed = "aed"
ad = "ad"
ae = "ae"
ed = "ed"
aedrad = "aedrad"
# arad = 'arad'
# erad = 'erad'

xyz = "xyz"
xy = "xy"
xz = "xz"
yz = "yz"

nxyz = "nxyz"
nxyzd = "nxyzd"
nxy = "nxy"
nxyd = "nxyd"
nxz = "nxz"
nxzd = "nxzd"
nyz = "nyz"
nyzd = "nyzd"

source_attributes = "sourceattributes"
doppler = "doppler"
plane = "planewave"
planewave = "planewave"
angle = "angle"


posformat = {
    # cartesian xyz
    x: (cartesian, (x,), False),
    y: (cartesian, (y,), False),
    z: (cartesian, (z,), False),
    xy: (cartesian, (x, y), False),
    xz: (cartesian, (x, z), False),
    yz: (cartesian, (y, z), False),
    xyz: (cartesian, (x, y, z), True),
    # polar aed
    azim: (polar, (azim,), False),
    a: (polar, (azim,), False),
    elev: (polar, (elev,), False),
    e: (polar, (elev,), False),
    dist: (polar, (dist,), False),
    d: (polar, (dist,), False),
    ad: (polar, (azim, dist), False),
    ed: (polar, (elev, dist), False),
    ae: (polar, (azim, elev), False),
    aed: (polar, (azim, elev, dist), True),
    # Polar in rad
    aedrad: (polar_rad, (az_rad, el_rad, dist), True),
    # oscar specific "normcartesian
    nx: (normcartesian, (nx,), False),
    nxd: (normcartesian, (nx, nd), False),
    ny: (normcartesian, (ny,), False),
    nyd: (normcartesian, (ny, nd), False),
    nz: (normcartesian, (nz,), False),
    nzd: (normcartesian, (nzd,), False),
    nxy: (normcartesian, (nx, ny), False),
    nxyd: (normcartesian, (nx, ny, nd), False),
    nyz: (normcartesian, (ny, nz), False),
    nyzd: (normcartesian, (ny, nz, nd), False),
    nxyz: (normcartesian, (nx, ny, nz), False),
    nxyzd: (normcartesian, (nx, ny, nz, nd), True),
    nd: (normcartesian, (nd,), False),
}
fullformat = {
    xyz: (x, y, z),
    nxyzd: (nx, ny, nz, nd),
    aed: (azim, elev, dist),
    cartesian: (x, y, z),
    normcartesian: (nx, ny, nz, nd),
    polar: (azim, elev, dist),
}
fullnames = {azim: "azimuth", elev: "elevation", dist: "distance"}

from enum import Enum
import numpy as np


class sControl_state(Enum):
    automation_control = np.ubyte(1)
    manually_control = np.ubyte(2)
    auto_switch_control = np.ubyte(0)


class OscPathType(Enum):
    Position = 1
    Properties = 2
    Gain = 3


osc_paths = {
    OscPathType.Position: {
        "base": [
            "/source/{val}",
            "/source/pos/{val}",
            "/source/position/{val}",
        ],
        "extended": [
            "/source/{idx}/{val}",
            "/source/{idx}/pos/{val}",
            "/source/{idx}/position/{val}",
        ],
    },
    OscPathType.Properties: {
        "base": ["/source/{val}"],
        "extended": ["/source/{idx}/{val}"],
    },
    OscPathType.Gain: {
        "base": [
            "/source/send/{val}",
            "/source/send/{val}/gain",
            "/send/{val}",
            "/send/{val}/gain",
            "/source/{val}",
            "/source/{val}/gain",
        ],
        "extended": [
            "/source/{idx}/send/{val}",
            "/source/{idx}/send/{val}/gain",
            "/send/{idx}/{val}",
            "/send/{idx}/{val}/gain",
            "/source/{idx}/{val}",
            "/source/{idx}/{val}/gain",
        ],
    },
}

osc_aliases = {
    "ambi": ["ambi", "ambisonics"],
    "wfs": ["wfs", "wavefieldsynthesis"],
    "reverb": ["reverb", "rev"],
}

knownAttributes = {planewave, doppler, angle}


class SourceAttributes(Enum):
    planewave = planewave
    doppler = doppler
    angle = angle


class CoordFormats(Enum):
    x = "x"
    y = "y"
    z = "z"
    azim = "azim"
    azimuth = "azim"
    elev = "elev"
    elevation = "elev"
    dist = "dist"
    distance = "dist"
    angle = "angle"

    nx = "nx"
    nxd = "nxd"
    ny = "ny"
    nyd = "nyd"
    nz = "nz"
    nzd = "nzd"
    nd = "dist"  # distance

    aed = "aed"
    aedrad = "aedrad"
    ad = "ad"
    ae = "ae"
    ed = "ed"
    xyz = "xyz"
    xy = "xy"
    xz = "xz"
    yz = "yz"

    nxyz = "nxyz"
    nxyzd = "nxyzd"
    nxy = "nxy"
    nxyd = "nxyd"
    nxz = "nxz"
    nxzd = "nxzd"
    nyz = "nyz"
    nyzd = "nyzd"


# global config keywords
globalconfig = "globalconfig"
inputport_data = "inputport_data"
inputport_ui = "inputport_ui"
inputport_settings = "inputport_settings"
number_sources = "number_sources"
max_gain = "max_gain"
send_changes_only = "send_changes_only"
data_port_timeout = "data_port_timeout"

renderengine = "renderengine"



import numpy as np
import osc_kreuz.str_keys_conventions as skc
import osc_kreuz.conversionsTools as ct
from functools import partial
from time import time

_tt = "time"
_uiBlock = "_uiDidBlock"


class SoundObject(object):

    globalConfig: dict = {}
    # data_port_timeout = 0.5
    number_renderer = 1
    # maxGain = 2.
    send_change_only = False
    #
    preferUi = True
    dataPortTimeOut = 1.0

    @classmethod
    def readGlobalConfig(cls, config: dict):
        cls.globalConfig = config
        cls.preferUi = bool(not config["data_port_timeout"] == 0)
        cls.dataPortTimeOut = float(config["data_port_timeout"])
        # cls.number_renderer =

    def __init__(self, objectID=0):

        self.objectID = objectID

        self._position: dict[str, np.float32] = {
            skc.x: ct.f32(0.0),
            skc.y: ct.f32(1.0),
            skc.z: ct.f32(0.0),
            skc.azim: ct.f32(0.0),
            skc.elev: ct.f32(0.0),
            skc.dist: ct.f32(1.0),
            skc.angle: ct.f32(0.0),
            skc.nx: ct.f32(0.0),
            skc.ny: ct.f32(1.0),
            skc.nz: ct.f32(0.0),
            skc.az_rad: ct.f32(0.0),
            skc.el_rad: ct.f32(0.0),
        }

        self._sourceattributes = {
            skc.SourceAttributes.planewave: 0,
            skc.SourceAttributes.doppler: 0,
            skc.SourceAttributes.angle: 0.0,
        }
        self._changedaAttributes: set = set()

        if "render_units" in self.globalConfig:
            self.number_renderer = len(self.globalConfig["render_units"])

        self._torendererSends = [np.float32(0.0) for _ in range(self.number_renderer)]
        self._changedSends: set = set()
        self._directSends = [0.0] * self.globalConfig["number_direct_sends"]
        self._changedDirSends = set()

        self._positionIsSet = {
            skc.polar: False,
            skc.cartesian: True,
            skc.normcartesian: True,
            skc.polar_rad: False,
        }
        self._lastUpdateKey = ""

        self._contState = skc.sControl_state.auto_switch_control

        self.uiBlockingDict = {}
        self.uiBlockingDict["position"] = self.createBlockingDict()
        #     {
        #     _tt: time(),
        #     _uiBlock: False
        # }
        # self.t_position = time()
        # self._uiDidSetPosition = False
        self.uiBlockingDict["attribute"] = self.createBlockingDict()
        # self.t_setAttribute = time()
        # self._uiDidSetAttribute = False
        self.uiBlockingDict["rendergain"] = []
        for i in range(self.number_renderer):
            self.uiBlockingDict["rendergain"].append(self.createBlockingDict())
        self.uiBlockingDict["directsend"] = []
        for i in range(self.globalConfig["number_direct_sends"]):
            self.uiBlockingDict["directsend"].append(self.createBlockingDict())

        # self.t_renderGain = [time()] * self.number_renderer

        # self._uiDidSetRenderGain = [False] * self.number_renderer
        # self.t_directSend = [time()] * self.globalConfig['number_direct_sends']
        # self._uiDidSetDirectSend = [False] * self.globalConfig['number_direct_sends']

    def createBlockingDict(self) -> dict:
        return {_tt: time(), _uiBlock: False}

    def _getPositionValuesForKey(self, key: skc.CoordFormats) -> list[np.float32]:
        vals = []
        for kk in skc.posformat[key.value][1]:
            vals.append(self._position[kk])

        return vals

    def setControlState(self, state: skc.sControl_state):
        self._contState = state

    def updateCoordinateFormat(self, updateFormat):

        # if updateFormat =
        _calcRad = False
        _calcDeg = False
        if updateFormat == skc.polar_rad:
            if self._positionIsSet[skc.polar]:
                _convert = False
            else:
                _convert = True
            updateFormat = skc.polar
            _calcRad = True

        elif updateFormat == skc.polar and self._positionIsSet[skc.polar_rad]:
            _calcDeg = True
            _convert = False

        else:
            _calcRad = False
            _convert = True

        if _convert:
            statusTupel = (
                self._positionIsSet[skc.polar],
                self._positionIsSet[skc.cartesian],
                self._positionIsSet[skc.normcartesian],
            )

            conversionMap = {
                skc.polar: {
                    (False, True, True): (ct.conv_ncart2pol, skc.CoordFormats.nxyzd),
                    (False, True, False): (ct.conv_cart2pol, skc.CoordFormats.xyz),
                    (False, False, True): (ct.conv_ncart2pol, skc.CoordFormats.nxyzd),
                },
                skc.cartesian: {
                    (True, False, True): (ct.conv_ncart2cart, skc.CoordFormats.nxyzd),
                    (True, False, False): (ct.conv_pol2cart, skc.CoordFormats.aed),
                    (False, False, True): (ct.conv_ncart2cart, skc.CoordFormats.nxyzd),
                },
                skc.normcartesian: {
                    (True, True, False): (ct.conv_cart2ncart, skc.CoordFormats.xyz),
                    (True, False, False): (ct.conv_pol2ncart, skc.CoordFormats.aed),
                    (False, True, False): (ct.conv_cart2ncart, skc.CoordFormats.xyz),
                },
            }
            convert_metadata = conversionMap[updateFormat][statusTupel]
            conversion = partial(
                convert_metadata[0]
            )  # , skc.posformat[convert_metadata[1]])

            updatedCoo = conversion(*self._getPositionValuesForKey(convert_metadata[1]))
            for idx, coo in enumerate(skc.fullformat[updateFormat]):
                self._position[coo] = updatedCoo[idx]

        if _calcRad:
            # self._position[skc.az_rad] = ct.deg2rad(self._position[skc.azim])
            # self._position[skc.el_rad] = ct.deg2rad(self._position[skc.elev])
            [self._position[skc.az_rad], self._position[skc.el_rad]] = (
                ct.convert_deg_angleToRad_correctMath(
                    self._position[skc.azim], self._position[skc.elev]
                )
            )
            self._positionIsSet[skc.polar_rad] = True
        elif _calcDeg:
            self._position[skc.azim] = ct.rad2deg(self._position[skc.az_rad])
            self._position[skc.elev] = ct.rad2deg(self._position[skc.el_rad])

        self._positionIsSet[updateFormat] = True

    def setPosition(self, coordinate_key: str, *values, fromUi: bool = True) -> bool:

        if not self.shouldProcessInput(self.uiBlockingDict["position"], fromUi):
            return False

        # if fromUi:
        #     self.gotUiInput(self.uiBlockDict['position'])
        #
        # elif self.preferUi and self.dataPortStillBlocked(self.uiBlockDict['position']):
        #     return False

        # if self.preferUi:
        #     if not fromUi and self._uiDidSetPosition:
        #         if self.dataPortStillBlocked(self.t_position):
        #             return False
        #         else:
        #             self._uiDidSetPosition = False
        #     else:
        #         self.t_position = time()
        #         self._uiDidSetPosition = True
        #
        coordinateType = skc.posformat[coordinate_key][0]
        sthChanged = False

        # XXX why are the ones with skc.posformat[coordinate_key][2]==False ignored? what is even happening here?
        if (
            not self._positionIsSet[coordinateType]
            and not skc.posformat[coordinate_key][2]
        ):
            self.updateCoordinateFormat(coordinateType)
            sthChanged = True

        for kk in self._positionIsSet.keys():
            self._positionIsSet[kk] = False

        self._lastUpdateKey = coordinate_key

        # TODO:copy old position

        # here the Position is set
        for idx, key in enumerate(skc.posformat[coordinate_key][1]):
            newValue = ct.f32(values[idx])
            if key == skc.dist:
                newValue = np.maximum(self.globalConfig["min_dist"], newValue)
            if self.globalConfig[skc.send_changes_only]:
                if not newValue == self._position[key]:
                    sthChanged = True
                    self._position[key] = newValue
            else:
                self._position[key] = newValue
                sthChanged = True

            # self._position[key] = ct.f32(values[idx])

        self._positionIsSet[coordinateType] = True

        # TODO: compare new Position with old position for return
        return sthChanged

    # def setPositionFromAutomation(self, coordinate_key: str, values) -> bool:
    #
    #     if self._contState == skc.sControl_state.automation_control:
    #         return self.setPosition(coordinate_key, values)
    #
    #     elif self._contState == skc.sControl_state.auto_switch_control:
    #         if self._uiDidSetRenderGain:
    #             if (time() - self._lastPositionUpdateTime) < self.globalConfig[skc.data_port_timeout]:
    #                 self._uiDidSetRenderGain = False
    #             else:
    #                 return False
    #
    #         return self.setPosition(coordinate_key, values)
    #     else:
    #         return False
    #
    #
    # def setPositionFromUserInterface(self, coordinate_key: str, values) -> bool:
    #     if self._contState == skc.sControl_state.auto_switch_control or self._contState == skc.sControl_state.manually_control:
    #         self._uiDidSetRenderGain = True
    #         self._lastPositionUpdateTime = time()
    #         self.setPosition(coordinate_key, values)
    #         return True
    #     else:
    #         return False

    def getSingleValueUpdate(self, keys):
        if self._lastUpdateKey in keys:
            return (self._lastUpdateKey, self._position[self._lastUpdateKey])
        else:
            return False

    def getPosition(self, pos_key: skc.CoordFormats) -> list[float]:
        """Get Position of this sound object in the format specified in pos_key

        Args:
            pos_key (str): str key of the desired position format

        Returns:
            list[float]: desired coordinates in a list. ATTENTION: If only one coordinate is requested, it is instead returned as a float.
        """
        coordinateType = skc.posformat[pos_key.value][0]

        if not self._positionIsSet[coordinateType]:
            self.updateCoordinateFormat(coordinateType)

        coords = []  # np.array([], dtype=np.float32)
        for key in skc.posformat[pos_key.value][1]:
            coords.append(float(self._position[key]))
        # float_coords = coords.
        if len(coords) == 1:
            return coords[0]
        else:
            return coords

    def setAttribute(self, attribute, value, fromUi: bool = True) -> bool:

        if not self.shouldProcessInput(self.uiBlockingDict["attribute"], fromUi):
            return False

        # if self.preferUi:
        #     if not fromUi and self._uiDidSetAttribute:
        #         if self.dataPortStillBlocked(self.t_setAttribute):
        #             return False
        #         else:
        #             self._uiDidSetAttribute = False
        #     else:
        #         self.t_position = time()
        #         self._uiDidSetAttribute = True

        if not self._sourceattributes[attribute] == value:
            self._sourceattributes[attribute] = value
            return True
        else:
            return False

    def getAttribute(self, attribute: skc.SourceAttributes):
        return self._sourceattributes[attribute]

    def setRendererGain(self, rendIdx: int, gain: float, fromUi: bool = True) -> bool:

        if not self.shouldProcessInput(
            self.uiBlockingDict["rendergain"][rendIdx], fromUi
        ):
            return False

        # if self.preferUi:
        #     if not fromUi and self._uiDidSetRenderGain[rendIdx]:
        #         if self.dataPortStillBlocked(self.t_renderGain[rendIdx]):
        #             return False
        #         else:
        #             self._uiDidSetRenderGain[rendIdx] = False
        #     else:
        #         self.t_position = time()
        #         self._uiDidSetRenderGain[rendIdx] = True

        _gain = ct.f32(np.clip(gain, a_min=0, a_max=self.globalConfig[skc.max_gain]))

        if self.globalConfig[skc.send_changes_only]:
            if self._torendererSends[rendIdx] == _gain:
                return False

        self._torendererSends[rendIdx] = _gain
        # self._changedSends.add(rendIdx)
        return True

    def setDirectSend(self, directIdx: int, gain: float, fromUi: bool = True) -> bool:
        # TODO: replace the fromUi thing with a function
        if not self.shouldProcessInput(
            self.uiBlockingDict["directsend"][directIdx], fromUi
        ):
            return False

        # if self.preferUi:
        #     if not fromUi and self._uiDidSetDirectSend[directIdx]:
        #         if self.dataPortStillBlocked(self.t_directSend[directIdx]):
        #             return False
        #         else:
        #             self._uiDidSetDirectSend[directIdx] = False
        #     else:
        #         self.t_position = time()
        #         self._uiDidSetDirectSend[directIdx] = True

        _gain = ct.f32(np.clip(gain, a_min=0, a_max=self.globalConfig[skc.max_gain]))

        if self.globalConfig[skc.send_changes_only]:
            if self._torendererSends[directIdx] == _gain:
                return False

        self._directSends[directIdx] = _gain
        # self._changedDirSends.add(directIdx)
        return True

    def checkIsBlocked(self, *args):
        pass

    def getBoolAndTime(self, *args):
        pass

    def getAllRendererGains(self) -> list[np.float32]:
        return self._torendererSends

    def getRenderGain(self, rIdx: int) -> float:
        return float(self._torendererSends[rIdx])

    def getAllDirectSends(self) -> list[float]:
        return self._directSends

    def getDirectSend(self, cIdx: int) -> float:
        return float(self._directSends[cIdx])

    def shouldProcessInput(
        self,
        blockDict: dict,
        fromUi: bool = True,
    ) -> bool:
        if fromUi:
            self.gotUiInput(blockDict)
            return True
        elif self.preferUi and self.dataPortStillBlocked(blockDict):
            return False
        else:
            return True

    def gotUiInput(self, blockDict: dict):
        blockDict[_uiBlock] = True
        blockDict[_tt] = time()

    def dataPortStillBlocked(self, blockDict: dict) -> bool:
        if blockDict[_uiBlock]:
            if time() - blockDict[_tt] > self.dataPortTimeOut:
                blockDict[_uiBlock] = False

        return blockDict[_uiBlock]

    # def getChangedDirectSends(self) -> [(int, float)]:
    #     msgs = []
    #     while self._changedDirSends:
    #         sendIdx = self._changedSends.pop()
    #         msgs.append((sendIdx, self._directSends[sendIdx]))
    #     return msgs
    #
    # def getChangedGains(self) -> ([],[]):
    #     return (self.getChangedRendererGains(), self.getChangedDirectSends())
    #
    # def getChangedRendererGains(self) -> [(int, float)]:
    #     msgs = []
    #     while self._changedSends:
    #         sendIdx = self._changedSends.pop()
    #         msgs.append((sendIdx, self._torendererSends[sendIdx]))
    #     return msgs
    # def getChangedAttributes(self, attribute, value):
    #     pass


# does nothing for now
class LinkedObject(object):
    def __init__(self, **kwargs):
        super(LinkedObject, self).__init__(**kwargs)



#!/usr/bin/env python3

import osc_kreuz.str_keys_conventions as skc

from osc_kreuz.soundobject import SoundObject
from osc_kreuz.renderer import Renderer
import osc_kreuz.renderer as rendererclass
import osc_kreuz.osccomcenter as osccomcenter

from pathlib import Path
from importlib.resources import files

import click
import signal
import yaml
import logging
import sys

logFormat = "%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]: %(message)s"
timeFormat = "%Y-%m-%d %H:%M:%S"
logging.basicConfig(format=logFormat, datefmt=timeFormat, level=logging.INFO)
log = logging.getLogger("main")


# lists for constructing default config paths
default_config_file_path = Path("osc-kreuz")
default_config_file_name_options = [
    "osc-kreuz_conf.yml",
    "osc-kreuz-conf.yml",
    "osc-kreuz_config.yml",
    "osc-kreuz-config.yml",
    "config.yml",
    "conf.yml",
]
default_config_file_locations = [
    Path.home() / ".config",
    Path("/etc"),
    Path("/usr/local/etc"),
]


def debug_prints(globalconfig, extendedOscInput, verbose):
    log.debug("max number of sources is set to %s", str(globalconfig["number_sources"]))
    log.debug("number of rendering units is %s", str(globalconfig["n_renderengines"]))

    try:
        render_units = globalconfig["render_units"]
    except KeyError:
        render_units = []

    for key in ["ambi", "wfs", "reverb"]:
        if key in render_units:
            log.debug(f"{key} index: {render_units.index(key)}")

    log.debug("UI listenport: %s", globalconfig[skc.inputport_ui])
    log.debug("DATA listenport (for automation): %s", globalconfig[skc.inputport_data])
    log.debug(
        "port for settings, ping and client subscription is: %s",
        globalconfig[skc.inputport_settings],
    )
    if extendedOscInput:
        log.debug("extended osc-string listening activated")
    else:
        log.debug("only basic osc-strings will be accepted")

    log.debug("max gain is %s", globalconfig[skc.max_gain])

    if Renderer.debugCopy:
        log.debug("Osc-Messages will be copied to somewhere")
    else:
        log.debug("No Debug client configured")

    log.debug("Verbosity Level is %s", verbose)
    if verbose == 1:
        log.debug("outgoing osc will be printed in console")
    elif verbose > 1:
        log.debug("incoming and outgoing osc will be printed in console")


@click.command(help="OSC Message Processor and Router")
@click.option(
    "-c",
    "--config",
    "config_path",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True, path_type=Path),
    help="path to configfile",
)
@click.option(
    "--oscdebug",
    help='ip and port for debug messages, e.g. "130.149.23.46:55112"',
    type=click.STRING,
)
@click.option("-v", "--verbose", count=True, help="increase verbosity level.")
@click.version_option()
def main(config_path, oscdebug, verbose):
    osccomcenter.setVerbosity(verbose)
    if verbose > 0:
        log.setLevel(logging.DEBUG)
    # get Config Path:
    if config_path is None:
        # check different paths for a config file, with the highest one taking precedence
        for possible_config_path in (
            base / default_config_file_path / filename
            for base in default_config_file_locations
            for filename in default_config_file_name_options
        ):
            if possible_config_path.exists():
                config_path = possible_config_path
                log.info(f"Loading config file {config_path}")
                break

    if config_path is None:
        log.warn(f"Could not find config, loading default config")
        config_path = files("osc_kreuz").joinpath("config_default.yml")
        config = yaml.load(config_path.read_bytes(), Loader=yaml.Loader)
    else:
        # read config file
        with open(config_path) as f:
            config = yaml.load(f, Loader=yaml.Loader)

    # setup debug osc client
    if oscdebug:
        oscDebugParams = oscdebug.split(":")
        debugIp = oscDebugParams[0]
        debugPort = int(oscDebugParams[1])
        Renderer.createDebugClient(debugIp, debugPort)
        Renderer.debugCopy = True

    # set extended OSC String Input Format (whatever that may be)
    extendedOscInput = True
    osccomcenter.extendedOscInput = extendedOscInput

    # prepare globalconfig
    globalconfig = config["globalconfig"]

    try:
        n_renderunits = len(globalconfig["render_units"])
    except KeyError:
        n_renderunits = 0
    globalconfig["n_renderengines"] = n_renderunits

    # set global config in objects
    SoundObject.readGlobalConfig(globalconfig)
    SoundObject.number_renderer = n_renderunits
    Renderer.globalConfig = globalconfig
    osccomcenter.globalconfig = globalconfig

    # setup number of sources
    numberofsources = int(globalconfig["number_sources"])  # 64

    Renderer.numberOfSources = numberofsources

    # Data initialisation
    soundobjects: list[SoundObject] = []
    for i in range(numberofsources):
        soundobjects.append(SoundObject(objectID=i + 1))

    # soundobjects are added as a class variable to the render class, so every renderer has access to them
    Renderer.sources = soundobjects

    receivers: list[Renderer] = []

    # creating audiorouters
    log.info("setting up receivers")
    if "receivers" in config:
        for receiver_config in config["receivers"]:
            if not "type" in receiver_config:
                log.warning("receiver has no type specified, skipping")
                continue
            try:
                receivers.append(rendererclass.createRendererClient(receiver_config))
            except rendererclass.RendererException as e:
                log.error(e)
                sys.exit(-1)

    # Setup OSC Com center
    osccomcenter.soundobjects = soundobjects
    osccomcenter.receivers = receivers

    osccomcenter.setupOscBindings()

    #
    if verbose > 0:
        debug_prints(globalconfig, extendedOscInput, verbose)

    log.info("OSC router ready to use")
    log.info("have fun...")

    signal.pause()


if __name__ == "__main__":
    main()



from numpy import iterable
from osc_kreuz.soundobject import SoundObject
import osc_kreuz.str_keys_conventions as skc
from oscpy.client import OSCClient

from threading import Timer
from functools import partial
from typing import Any, Callable

from collections.abc import Iterable
import logging

log = logging.getLogger("renderer")
verbosity = 0


class RendererException(Exception):
    pass


class Message:
    def __init__(self, path: bytes, values: Any) -> None:
        self.path = path
        if isinstance(values, str) or not iterable(values):
            self.values = [values]
        self.values: Iterable[Any] = values


class Update:
    def __init__(
        self,
        path: bytes,
        soundobject: SoundObject,
        source_index: int | None = None,
        pre_arg: Any = None,
        post_arg: Any = None,
    ):
        self.soundobject = soundobject
        self.pre_arg = pre_arg
        self.post_arg = post_arg
        self.path = path
        self.source_index = source_index

    def get_value(self):
        raise NotImplementedError

    def __eq__(self, other):
        """Overrides the default implementation"""
        if isinstance(other, self.__class__):
            return (
                isinstance(other, self.__class__)
                and self.source_index == other.source_index
                and self.path == other.path
            )
        return False

    def __hash__(self):
        """for use with sets"""
        return hash(tuple(sorted(self.__dict__.items())))

    def to_message(self) -> Message:
        values = []

        # add source index to list if it exists
        if self.source_index is not None:
            values.append(self.source_index)

        # sometimes needed as first argument
        if self.pre_arg is not None:
            values.append(self.pre_arg)

        # add value or values returned by the callback to the list
        ret_value = self.get_value()
        if isinstance(ret_value, str) or not iterable(ret_value):
            values.append(ret_value)
        else:
            values.extend(ret_value)

        # sometimes needed as a last argument
        if self.post_arg is not None:
            values.append(self.post_arg)

        return Message(self.path, values)


class PositionUpdate(Update):
    def __init__(
        self,
        path: bytes,
        soundobject: SoundObject,
        coord_fmt: skc.CoordFormats,
        source_index: int | None = None,
        pre_arg: Any = None,
        post_arg: Any = None,
    ):
        super().__init__(path, soundobject, source_index, pre_arg, post_arg)
        self.coord_fmt = coord_fmt

    def get_value(self):
        return self.soundobject.getPosition(self.coord_fmt)


class GainUpdate(Update):
    def __init__(
        self,
        path: bytes,
        soundobject: SoundObject,
        render_idx: int,
        source_index: int | None = None,
        pre_arg: Any = None,
        post_arg: Any = None,
        include_render_idx=False,
    ):
        super().__init__(path, soundobject, source_index, pre_arg, post_arg)
        self.render_idx = render_idx
        if include_render_idx:
            self.pre_arg = render_idx

    def get_value(self):
        return self.soundobject.getRenderGain(self.render_idx)


class DirectSendUpdate(Update):
    def __init__(
        self,
        path: bytes,
        soundobject: SoundObject,
        send_index: int,
        source_index: int | None = None,
        include_send_idx=False,
        pre_arg: Any = None,
        post_arg: Any = None,
    ):
        super().__init__(path, soundobject, source_index, pre_arg, post_arg)
        self.send_index = send_index
        if include_send_idx:
            self.pre_arg = send_index

    def get_value(self):
        return self.soundobject.getDirectSend(self.send_index)


class AttributeUpdate(Update):
    def __init__(
        self,
        path: bytes,
        attribute: skc.SourceAttributes,
        soundobject: SoundObject,
        source_index: int | None = None,
        pre_arg: Any = None,
        post_arg: Any = None,
    ):
        super().__init__(path, soundobject, source_index, pre_arg, post_arg)
        self.attribute = attribute

    def get_value(self):
        return self.soundobject.getAttribute(self.attribute)


class wonderPlanewaveAttributeUpdate(AttributeUpdate):
    def get_value(self):
        # for the planewave attribute, the value has to be inverted
        return int(not super().get_value())


class Renderer(object):

    numberOfSources = 64
    sources: list[SoundObject] = []
    globalConfig: dict = {}
    debugCopy: bool = False
    oscDebugClient: OSCClient

    printOutput = verbosity >= 1

    @classmethod
    def createDebugClient(cls, ip, port):
        cls.oscDebugClient = OSCClient(ip, port)

    @classmethod
    def setVerbosity(cls, v: int):
        global verbosity
        verbosity = v
        cls.printOutput = verbosity >= 1

    def __init__(
        self,
        dataformat: skc.CoordFormats | str = skc.CoordFormats.xyz,
        updateintervall=10,
        hostname="127.0.0.1",
        hosts: list[dict] | None = None,
        port=4002,
        sourceattributes=(),
        indexAsValue=0,  # XXX unused
    ):
        self.setVerbosity(verbosity)

        self.posFormat = skc.CoordFormats(dataformat)
        self.validSinglePosKeys = {}
        self.sourceAttributes = sourceattributes

        # check if hosts are defined as an array
        self.hosts: list[tuple[str, int]] = []
        if hosts is None:
            self.hosts.append((hostname, int(port)))
        else:
            for host in hosts:
                try:
                    host_tuple = (host["hostname"], host["port"])
                    self.hosts.append(host_tuple)
                except KeyError:
                    raise RendererException("Invalid Host")

        self.updateIntervall = int(updateintervall) / 1000

        self.source_needs_update: list[bool] = [False] * self.numberOfSources
        self.source_getting_update: list[bool] = [False] * self.numberOfSources

        # update stack contains sets of tuples of functions and osc_paths
        # sets are used, so each source is updated only once during the update process
        self.updateStack: list[set[Update]] = [
            set() for _ in range(self.numberOfSources)
        ]

        self.debugPrefix = "/genericRenderer"
        self.oscPre = ("/source/" + self.posFormat.value).encode()

        self.receivers: list[OSCClient] = []
        for ip, port in self.hosts:
            self.receivers.append(OSCClient(ip, port, encoding="utf8"))

        self.isDataClient = False

        self.printRenderInformation()

    def printRenderInformation(self, print_pos_format=True):
        log.info(f"Initialized renderer {self.myType()}")
        hosts_str = ", ".join([f"{hostname}:{port}" for hostname, port in self.hosts])
        log.info(f"\thosts: {hosts_str}")
        if print_pos_format:
            log.info(f"\tlistening to format {self.posFormat}")

    def myType(self) -> str:
        return "basic Rendererclass: abstract class, doesnt listen"

    def addDestination(self, ip: str, port: int):
        self.receivers.append(OSCClient(ip, port, encoding="utf8"))

        log.debug(self.myType(), "added destination", ip, str(port))

    def sourceChanged(self, source_idx):
        if not self.source_getting_update[source_idx]:
            self.updateSource(source_idx)
        else:
            self.source_needs_update[source_idx] = True

    def add_update(self, source_idx: int, update: Update) -> None:
        self.updateStack[source_idx].add(update)
        self.sourceChanged(source_idx)

    def updateSource(self, source_idx):
        """Builds and sends source update messages

        Args:
            source_idx (int): index of source to be updated
        """
        while self.updateStack[source_idx]:
            update: Update = self.updateStack[source_idx].pop()
            msg = update.to_message()
            self.sendUpdates([msg])

        self.scheduleSourceUpdateCheck(source_idx)

    def sendUpdates(self, msgs):
        """This function sends all messages to the osc clients

        Args:
            msgs (list(list)): list of messages
        """
        for msg in msgs:
            for receiversClient in self.receivers:
                try:
                    receiversClient.send_message(msg.path, msg.values)

                except Exception as e:
                    log.warn(f"Exception while sending: {e}")
                    pass

                if self.debugCopy:
                    debugOsc = (
                        self.debugPrefix
                        + "/"
                        + receiversClient.address
                        + ":"
                        + str(receiversClient.port)
                        + msg.path.decode()
                    ).encode()
                    try:
                        self.oscDebugClient.send_message(debugOsc, msg.values)
                    except:
                        pass

            if self.printOutput:
                self.printOscOutput(msg.path, msg.values)

    def scheduleSourceUpdateCheck(self, source_idx):
        self.source_needs_update[source_idx] = False
        self.source_getting_update[source_idx] = True
        Timer(
            self.updateIntervall, partial(self.check_sourceNeedsUpdate, source_idx)
        ).start()

    def check_sourceNeedsUpdate(self, source_idx):
        self.source_getting_update[source_idx] = False
        if self.source_needs_update[source_idx]:
            self.updateSource(source_idx)

    # implement this functions in subclasses for registering for specific updates
    def sourceAttributeChanged(self, source_idx, attribute):
        pass

    def sourceRenderGainChanged(self, source_idx, render_idx):
        pass

    def sourceDirectSendChanged(self, source_idx, send_idx):
        pass

    def sourcePositionChanged(self, source_idx):
        pass

    def oscDebugSend(self, oscStr, data: list):
        decStr = oscStr.decode()
        newOscAddr = self.debugPrefix + decStr
        self.oscDebugClient.send_message(newOscAddr.encode(), data)

    def printOscOutput(self, oscStr, data: list):
        decStr = oscStr.decode()
        log.debug("OSC out", self.debugPrefix, decStr, data)


class SpatialRenderer(Renderer):
    def myType(self) -> str:
        return "Generic Spatial Renderer"

    def sourcePositionChanged(self, source_idx):
        self.add_update(
            source_idx,
            PositionUpdate(
                path=self.oscPre,
                soundobject=self.sources[source_idx],
                coord_fmt=self.posFormat,
                source_index=source_idx,
            ),
        )


class Wonder(SpatialRenderer):
    def __init__(self, **kwargs):
        if not "dataformat" in kwargs.keys():
            kwargs["dataformat"] = skc.CoordFormats.xy
        if not "sourceattributes" in kwargs.keys():
            kwargs["sourceattributes"] = (
                skc.SourceAttributes.doppler,
                skc.SourceAttributes.planewave,
            )
        super(Wonder, self).__init__(**kwargs)

        self.attributeOsc = {
            skc.SourceAttributes.doppler: b"/WONDER/source/dopplerEffect",
            skc.SourceAttributes.planewave: b"/WONDER/source/type",
            skc.SourceAttributes.angle: b"/WONDER/source/angle",
        }
        self.oscPre = b"/WONDER/source/position"

        self.interpolTime = self.updateIntervall
        self.linkPositionAndAngle = True

        self.debugPrefix = "/dWonder"

    def myType(self) -> str:
        return "Wonder"

    def sourcePositionChanged(self, source_idx):
        # Add position Update to update stack
        self.add_update(
            source_idx,
            PositionUpdate(
                path=self.oscPre,
                soundobject=self.sources[source_idx],
                coord_fmt=self.posFormat,
                source_index=source_idx,
                post_arg=self.interpolTime,
            ),
        )

        # optionally update angle if the wave is planar
        if self.linkPositionAndAngle and self.sources[source_idx].getAttribute(
            skc.SourceAttributes.planewave
        ):
            self.sourceAttributeChanged(source_idx, skc.SourceAttributes.angle)

    def sourceAttributeChanged(self, source_idx, attribute: skc.SourceAttributes):
        if attribute == skc.SourceAttributes.planewave:
            # planewave has special update type
            self.add_update(
                source_idx,
                wonderPlanewaveAttributeUpdate(
                    path=self.attributeOsc[attribute],
                    soundobject=self.sources[source_idx],
                    source_index=source_idx,
                    attribute=attribute,
                ),
            )

            if self.sources[source_idx].getAttribute(attribute):
                self.update_auto_angle(source_idx)
        elif attribute == skc.SourceAttributes.angle:
            # angle needs interpolation time as additional param
            self.add_update(
                source_idx,
                AttributeUpdate(
                    path=self.attributeOsc[attribute],
                    soundobject=self.sources[source_idx],
                    source_index=source_idx,
                    attribute=attribute,
                    post_arg=self.interpolTime,
                ),
            )
        else:
            self.add_update(
                source_idx,
                AttributeUpdate(
                    path=self.attributeOsc[attribute],
                    soundobject=self.sources[source_idx],
                    source_index=source_idx,
                    attribute=attribute,
                ),
            )

    def update_auto_angle(self, source_idx: int):
        # TODO take into account the user specified angle
        self.add_update(
            source_idx,
            PositionUpdate(
                path=self.attributeOsc[skc.SourceAttributes.angle],
                soundobject=self.sources[source_idx],
                source_index=source_idx,
                coord_fmt=skc.CoordFormats.azim,
                post_arg=self.interpolTime,
            ),
        )


class Audiorouter(Renderer):
    #
    def __init__(self, **kwargs):
        super(Audiorouter, self).__init__(**kwargs)

        self.debugPrefix = "/dAudiorouter"
        self.oscpre_renderGain = b"/source/send/spatial"
        self.oscpre_reverbGain = b"/source/reverb/gain"
        self.oscpre_directSend = b"/source/send/direct"

    def printRenderInformation(self, print_pos_format=False):
        super().printRenderInformation(print_pos_format=print_pos_format)

    def myType(self) -> str:
        return "Audiorouter"

    def sourceDirectSendChanged(self, source_idx, send_idx):
        self.add_update(
            source_idx,
            DirectSendUpdate(
                self.oscpre_directSend,
                soundobject=self.sources[source_idx],
                send_index=send_idx,
                source_index=source_idx,
                include_send_idx=True,
            ),
        )

    def sourceRenderGainChanged(self, source_idx, render_idx):
        if render_idx == 1:
            return

        if render_idx == 2:
            path = self.oscpre_reverbGain
            include_render_idx = False
        else:
            path = self.oscpre_renderGain
            include_render_idx = True

        self.add_update(
            source_idx,
            GainUpdate(
                path=path,
                soundobject=self.sources[source_idx],
                render_idx=render_idx,
                source_index=source_idx,
                include_render_idx=include_render_idx,
            ),
        )


class AudiorouterWFS(Audiorouter):
    def __init__(self, **kwargs):
        super(AudiorouterWFS, self).__init__(**kwargs)
        self.debugPrefix = "/dAudiorouterWFS"

    def sourceRenderGainChanged(self, source_idx, render_idx):
        if render_idx != 1:
            return

        self.add_update(
            source_idx,
            GainUpdate(
                path=self.oscpre_renderGain,
                soundobject=self.sources[source_idx],
                render_idx=render_idx,
                source_index=source_idx,
                include_render_idx=True,
            ),
        )

    def myType(self) -> str:
        return "Audiorouter-WFS"


class AudioMatrix(Renderer):
    def __init__(self, paths: Iterable[dict["str", Any]], **kwargs):
        super().__init__(**kwargs)
        self.debugPrefix = "/dAudioMatrix"
        self.gain_paths: dict[int, list[bytes]] = {}
        self.pos_paths: list[tuple[bytes, skc.CoordFormats]] = []

        # this dict is used to translate between render unit index and render unit name
        self.render_unit_indices = {}

        # prepare gain path with all render unit indices
        for index, render_unit in enumerate(self.globalConfig["render_units"]):
            self.render_unit_indices[index] = render_unit
            self.render_unit_indices[render_unit] = index
            self.gain_paths[index] = []

        # add all configured paths from the yaml file to either the correct gain path index,
        # or to the position_path list
        for path in paths:
            osc_path: str = path["path"]
            path_type = path["type"]

            if path_type == "gain":
                renderer = path["renderer"]
                renderer_index = self.render_unit_indices[renderer]
                self.gain_paths[renderer_index].append(osc_path.encode())
            elif path_type in ["position", "pos"]:
                try:
                    coord_fmt = skc.CoordFormats(path["format"])
                except:
                    coord_fmt = skc.CoordFormats("xyz")
                self.pos_paths.append((osc_path.encode(), coord_fmt))

        log.debug("Audio Matrix initialized")

    def myType(self) -> str:
        return "AudioMatrix"

    def sourceRenderGainChanged(self, source_idx, render_idx):
        if render_idx in self.gain_paths:
            for path in self.gain_paths[render_idx]:
                self.add_update(
                    source_idx,
                    GainUpdate(
                        path=path,
                        soundobject=self.sources[source_idx],
                        render_idx=render_idx,
                        source_index=source_idx,
                    ),
                )

    def sourcePositionChanged(self, source_idx):
        for path, coord_fmt in self.pos_paths:
            self.add_update(
                source_idx,
                PositionUpdate(
                    path=path,
                    soundobject=self.sources[source_idx],
                    coord_fmt=coord_fmt,
                    source_index=source_idx,
                ),
            )


# class Panoramix(SpatialRenderer):
#     def __init__(self, **kwargs):
#         if not "dataformat" in kwargs.keys():
#             kwargs["dataformat"] = skc.xyz
#         super(Panoramix, self).__init__(**kwargs)

#         self.posAddrs = []
#         for i in range(self.numberOfSources):
#             self.posAddrs.append(("/track/" + str(i + 1) + "/xyz").encode())

#         self.debugPrefix = "/dPanoramix"

#     def myType(self) -> str:
#         return "Panoramix CAREFUL NOT REALLY IMPLEMENTED"

#     def composeSourceUpdateMessage(
#         self, values, sIdx: int = 0, *args
#     ) -> list[tuple[bytes, Iterable]]:
#         # msgs = []
#         sobject = self.sources[sIdx]
#         position = sobject.getPosition(self.posFormat)
#         # sourceID = source_idx + 1
#         addr = self.posAddrs[sIdx]

#         return [(addr, position)]


class SuperColliderEngine(SpatialRenderer):
    def __init__(self, **kwargs):
        if not "dataformat" in kwargs.keys():
            kwargs["dataformat"] = skc.aed
        super(SuperColliderEngine, self).__init__(**kwargs)

        self.oscPre = b"/source/pos/aed"

        self.debugPrefix = "/dSuperCollider"

    def myType(self) -> str:
        return "Supercolliderengine"


class ViewClient(SpatialRenderer):
    def myType(self) -> str:
        return "viewClient: {}".format(self.alias)

    def __init__(self, aliasname, **kwargs):
        self.alias = aliasname

        super(ViewClient, self).__init__(**kwargs)

        self.pingCounter = 0

        self.debugPrefix = "/d{}".format(aliasname.decode())
        # self.biAlias = b''
        # self.setAlias(aliasname)

        self.indexAsValue = False
        if "indexAsValue" in kwargs.keys():
            self.indexAsValue = kwargs["indexAsValue"]

        self.idxSourceOscPrePos = [b""] * self.numberOfSources
        self.idxSourceOscPreAttri = [{}] * self.numberOfSources
        self.idxSourceOscPreRender = [[]] * self.numberOfSources

        self.createOscPrefixes()

        # self.idxSourceOscPreAttri

        self.pingTimer: Timer | None = None

    def createOscPrefixes(self):
        for i in range(self.numberOfSources):
            self.idxSourceOscPrePos[i] = "/source/{}/{}".format(
                i + 1, self.posFormat.value
            ).encode()
            _aDic = {}
            for attr in skc.knownAttributes:
                _aDic[attr] = "/source/{}/{}".format(i + 1, attr).encode()

            self.idxSourceOscPreAttri[i] = _aDic

            try:
                render_units = self.globalConfig["render_units"]
            except KeyError:
                render_units = []

            renderList = [b""] * self.globalConfig["n_renderengines"]
            if (
                "ambi" in render_units
                and "wfs" in render_units
                and "reverb" in render_units
            ):
                renderList[render_units.index("ambi")] = "/source/{}/ambi".format(
                    i + 1
                ).encode()
                renderList[render_units.index("wfs")] = "/source/{}/wfs".format(
                    i + 1
                ).encode()
                renderList[render_units.index("reverb")] = "/source/{}/reverb".format(
                    i + 1
                ).encode()
            else:
                for j in range(self.globalConfig["number_renderunits"]):
                    self.idxSourceOscPreRender[i][j] = "/source/{}/send/{}".format(
                        i + 1, j
                    ).encode()
            self.idxSourceOscPreRender[i] = renderList

    def checkAlive(self, deleteClient):
        self.pingTimer = Timer(2.0, partial(self.checkAlive, deleteClient))

        if self.pingCounter < 6:
            # self.receivers[0].send_message(b'/oscrouter/ping', [self.globalConfig['inputport_settings']])
            try:
                self.receivers[0].send_message(
                    b"/oscrouter/ping", [self.globalConfig["inputport_settings"]]
                )  # , self.alias
            except:
                log.warn("error while pinging client ", self.alias, ", removing")
                self.pingTimer.cancel()
                deleteClient(self, self.alias)

            self.pingCounter += 1
            self.pingTimer.start()
        else:
            deleteClient(self, self.alias)

    def receivedIsAlive(self):
        self.pingCounter = 0

    def sourcePositionChanged(self, source_idx):
        if self.indexAsValue:
            path = self.idxSourceOscPrePos[source_idx]
            source_index_for_update = None
        else:
            path = self.oscPre
            source_index_for_update = source_idx
        self.add_update(
            source_idx,
            PositionUpdate(
                path=path,
                soundobject=self.sources[source_idx],
                coord_fmt=self.posFormat,
                source_index=source_index_for_update,
            ),
        )

    def sourceRenderGainChanged(self, source_idx, render_idx):
        self.add_update(
            source_idx,
            GainUpdate(
                self.idxSourceOscPreRender[source_idx][render_idx],
                soundobject=self.sources[source_idx],
                render_idx=render_idx,
            ),
        )


class Oscar(SpatialRenderer):
    def __init__(self, **kwargs):
        if not "dataformat" in kwargs.keys():
            kwargs["dataformat"] = skc.aed
        super(Oscar, self).__init__(**kwargs)

        self.sourceAttributes = (
            skc.SourceAttributes.doppler,
            skc.SourceAttributes.planewave,
        )

        # self.posAddrs = []

        self.oscPosPre = []
        self.oscAttrPre = []
        self.oscRenderPre = []
        self.oscDirectPre = []

        for i in range(self.numberOfSources):
            sourceAddrs = {}
            for kk in skc.fullformat[self.posFormat.value]:
                addrStr = "/source/" + str(i + 1) + "/" + kk
                sourceAddrs[kk] = addrStr.encode()
            self.oscPosPre.append(sourceAddrs)

            attrDic = {}
            for key in self.sourceAttributes:
                oscStr = "/source" + str(i + 1) + "/" + key.value
                attrDic[key] = oscStr.encode()
            self.oscAttrPre.append(attrDic)

            renderGainOscs = []
            for rId in range(self.globalConfig["n_renderengines"]):
                riOsc = "/source/" + str(i + 1) + "/render/" + str(rId)
                renderGainOscs.append(riOsc.encode())
            self.oscRenderPre.append(renderGainOscs)

            channelSend = []
            for cId in range(self.globalConfig["number_direct_sends"]):
                csOsc = "/source/" + str(i + 1) + "/direct/" + str(cId)
                channelSend.append(csOsc.encode())
            self.oscDirectPre.append(channelSend)

            # self.posAddrs.append(sourceAddrs)

        self.validPosKeys = {skc.dist}

        self.isDataClient = True

        self.debugPrefix = "/dOscar"

    def myType(self) -> str:
        return "Oscar"

    def sourcePositionChanged(self, source_idx):
        for key in skc.fullformat[self.posFormat.value]:
            self.add_update(
                source_idx,
                PositionUpdate(
                    self.oscPosPre[source_idx][key],
                    soundobject=self.sources[source_idx],
                    coord_fmt=skc.CoordFormats(key),
                ),
            )

    def sourceAttributeChanged(self, source_idx, attribute):
        self.add_update(
            source_idx,
            AttributeUpdate(
                path=self.oscAttrPre[source_idx][attribute],
                soundobject=self.sources[source_idx],
                attribute=attribute,
            ),
        )

    def sourceDirectSendChanged(self, source_idx, send_idx):
        self.add_update(
            source_idx,
            DirectSendUpdate(
                path=self.oscDirectPre[source_idx][send_idx],
                soundobject=self.sources[source_idx],
                send_index=send_idx,
            ),
        )

    def sourceRenderGainChanged(self, source_idx, render_idx):
        self.add_update(
            source_idx,
            GainUpdate(
                path=self.oscRenderPre[source_idx][render_idx],
                soundobject=self.sources[source_idx],
                render_idx=render_idx,
            ),
        )


class SeamlessPlugin(SpatialRenderer):
    def myType(self) -> str:
        return "Seamless Plugin"

    def __init__(self, **kwargs):
        if not "dataformat" in kwargs.keys():
            kwargs["dataformat"] = skc.xyz
        super(SeamlessPlugin, self).__init__(**kwargs)

        self.sourceAttributes = (
            skc.SourceAttributes.doppler,
            skc.SourceAttributes.planewave,
        )

        self.oscAddrs: dict = {}

        for key in skc.fullformat[self.posFormat.value]:
            self.oscAddrs[key] = "/source/pos/{}".format(key).encode()

        for vv in self.sourceAttributes:
            self.oscAddrs[vv.value] = "/{}".format(vv.value).encode()

        self.oscAddrs["renderGain"] = "/send/gain".encode()

        self.debugPrefix = "/dSeamlessPlugin"

    def composeSourceUpdateMessage(
        self, values, sIdx: int = 0, *args
    ) -> list[tuple[bytes, Iterable]]:
        osc_pre = args[0]
        if osc_pre == self.oscAddrs["renderGain"]:
            return [(osc_pre, [sIdx + 1, args[1], values])]
        else:
            return [(osc_pre, [sIdx + 1, values])]

    def sourceAttributeChanged(self, source_idx, attribute):
        pass

    def sourceRenderGainChanged(self, source_idx, render_idx):
        self.add_update(
            source_idx,
            GainUpdate(
                self.oscAddrs["renderGain"],
                soundobject=self.sources[source_idx],
                render_idx=render_idx,
                source_index=source_idx + 1,
                include_render_idx=True,
            ),
        )

    def sourcePositionChanged(self, source_idx):
        for key in skc.fullformat[self.posFormat.value]:
            self.add_update(
                source_idx,
                PositionUpdate(
                    path=self.oscAddrs[key],
                    soundobject=self.sources[source_idx],
                    coord_fmt=skc.CoordFormats(key),
                    source_index=source_idx + 1,
                ),
            )


class DataClient(Audiorouter, SpatialRenderer):
    pass


renderer_name_dict = {
    "wonder": Wonder,
    # "panoramix": Panoramix,
    "viewclient": ViewClient,
    "oscar": Oscar,
    "scengine": SuperColliderEngine,
    "audiorouter": Audiorouter,
    "seamlessplugin": SeamlessPlugin,
    "audiorouterwfs": AudiorouterWFS,
    "audiomatrix": AudioMatrix,
}


def createRendererClient(config: dict) -> Renderer:

    # XXX some weird shit is happening here
    if "dataformat" in config:
        tmp_dataFormat = config["dataformat"]
        if not tmp_dataFormat in skc.posformat.keys():
            if len(tmp_dataFormat.split("_")) == 2:
                preStr = ""
                if tmp_dataFormat.split("_")[0] == "normcartesian":
                    preStr = "n"

                dFo = preStr + tmp_dataFormat.split("_")[1]
                config["dataformat"] = dFo
            else:
                log.warn("unknown position format")
                del config["dataformat"]

    if "type" not in config:
        raise RendererException("Type of receiver unspecified")

    renderer_type = config["type"].lower()
    del config["type"]

    if renderer_type not in renderer_name_dict:
        raise RendererException(f"Invalid receiver type: {renderer_type}")

    return renderer_name_dict[renderer_type](**config)
